"""Tests for the player model class."""
import sys
import unittest

from sportsball.data.player_model import PlayerModel, VERSION
from sportsball.data.species import Species


class TestPlayerModel(unittest.TestCase):

    def setUp(self):
        sys.setrecursionlimit(2048)
        self._player_model = PlayerModel(
            identifier="a",
            jersey="45",
            kicks=None,
            fumbles=None,
            fumbles_lost=None,
            field_goals=None,
            field_goals_attempted=10,
            offensive_rebounds=None,
            assists=None,
            turnovers=None,
            name="James Jones",
            marks=None,
            handballs=None,
            disposals=None,
            goals=None,
            behinds=None,
            hit_outs=None,
            tackles=None,
            rebounds=None,
            insides=None,
            clearances=None,
            clangers=None,
            free_kicks_for=None,
            free_kicks_against=None,
            brownlow_votes=None,
            contested_possessions=None,
            uncontested_possessions=None,
            contested_marks=None,
            marks_inside=None,
            one_percenters=None,
            bounces=None,
            goal_assists=None,
            percentage_played=None,
            birth_date=None,
            age=None,
            species=str(Species.HUMAN),
            handicap_weight=None,
            father=None,
            starting_position=None,
            weight=None,
            birth_address=None,
            owner=None,
            seconds_played=None,
            three_point_field_goals=None,
            three_point_field_goals_attempted=None,
            free_throws=None,
            free_throws_attempted=None,
            defensive_rebounds=None,
            steals=None,
            blocks=None,
            personal_fouls=None,
            points=None,
            game_score=None,
            point_differential=None,
            version=VERSION,
            height=None,
            colleges=[],
            headshot=None,
            forced_fumbles=None,
            fumbles_recovered=None,
            fumbles_recovered_yards=None,
            fumbles_touchdowns=None,
            offensive_two_point_returns=None,
            offensive_fumbles_touchdowns=None,
            defensive_fumbles_touchdowns=None,
            average_gain=None,
            completion_percentage=None,
            completions=None,
            espn_quarterback_rating=None,
            interception_percentage=None,
            interceptions=None,
            long_passing=None,
            misc_yards=None,
            net_passing_yards=None,
            net_total_yards=None,
            passing_attempts=None,
            passing_big_plays=None,
            passing_first_downs=None,
            passing_fumbles=None,
            passing_fumbles_lost=None,
            passing_touchdown_percentage=None,
            passing_touchdowns=None,
            passing_yards=None,
            passing_yards_after_catch=None,
            quarterback_rating=None,
            sacks=None,
            passing_yards_at_catch=None,
            sacks_yards_lost=None,
            net_passing_attempts=None,
            total_offensive_plays=None,
            total_points=None,
            total_touchdowns=None,
            total_yards=None,
            total_yards_from_scrimmage=None,
            two_point_pass=None,
            two_point_pass_attempt=None,
            yards_per_completion=None,
            yards_per_pass_attempt=None,
            net_yards_per_pass_attempt=None,
            long_rushing=None,
            rushing_attempts=None,
            rushing_big_plays=None,
            rushing_first_downs=None,
            rushing_fumbles=None,
            rushing_fumbles_lost=None,
            rushing_touchdowns=None,
            rushing_yards=None,
            stuffs=None,
            stuff_yards_lost=None,
            two_point_rush=None,
            two_point_rush_attempts=None,
            yards_per_rush_attempt=None,
            espn_widereceiver=None,
            long_reception=None,
            receiving_big_plays=None,
            receiving_first_downs=None,
            receiving_fumbles=None,
            receiving_fumbles_lost=None,
            receiving_targets=None,
            receiving_touchdowns=None,
            receiving_yards=None,
            receiving_yards_after_catch=None,
            receiving_yards_at_catch=None,
            receptions=None,
            two_point_receptions=None,
            two_point_reception_attempts=None,
            yards_per_reception=None,
            assist_tackles=None,
            average_interception_yards=None,
            average_sack_yards=None,
            average_stuff_yards=None,
            blocked_field_goal_touchdowns=None,
            blocked_punt_touchdowns=None,
            defensive_touchdowns=None,
            hurries=None,
            kicks_blocked=None,
            long_interception=None,
            misc_touchdowns=None,
            passes_batted_down=None,
            passes_defended=None,
            quarterback_hits=None,
            sacks_assisted=None,
            sacks_unassisted=None,
            sacks_yards=None,
            safeties=None,
            solo_tackles=None,
            stuff_yards=None,
            tackles_for_loss=None,
            tackles_yards_lost=None,
            yards_allowed=None,
            points_allowed=None,
            one_point_safeties_made=None,
            missed_field_goal_return_td=None,
            blocked_punt_ez_rec_td=None,
            interception_touchdowns=None,
            interception_yards=None,
            average_kickoff_return_yards=None,
            average_kickoff_yards=None,
            extra_point_attempts=None,
            extra_point_percentage=None,
            extra_point_blocked=None,
            extra_points_blocked_percentage=None,
            extra_points_made=None,
            fair_catches=None,
            fair_catch_percentage=None,
            field_goal_attempts_max_19_yards=None,
            field_goal_attempts_max_29_yards=None,
            field_goal_attempts_max_39_yards=None,
            field_goal_attempts_max_49_yards=None,
            field_goal_attempts_max_59_yards=None,
            field_goal_attempts_max_99_yards=None,
            field_goal_attempts_above_50_yards=None,
            field_goal_attempt_yards=None,
            field_goals_blocked=None,
            field_goals_blocked_percentage=None,
            field_goals_made=None,
            field_goals_made_max_19_yards=None,
            field_goals_made_max_29_yards=None,
            field_goals_made_max_39_yards=None,
            field_goals_made_max_49_yards=None,
            field_goals_made_max_59_yards=None,
            field_goals_made_max_99_yards=None,
            field_goals_made_above_50_yards=None,
            field_goals_made_yards=None,
            field_goals_missed_yards=None,
            kickoff_out_of_bounds=None,
            kickoff_returns=None,
            kickoff_returns_touchdowns=None,
            kickoff_return_yards=None,
            kickoffs=None,
            kickoff_yards=None,
            long_field_goal_attempt=None,
            long_field_goal_made=None,
            long_kickoff=None,
            total_kicking_points=None,
            touchback_percentage=None,
            touchbacks=None,
            defensive_fumble_returns=None,
            defensive_fumble_return_yards=None,
            fumble_recoveries=None,
            fumble_recovery_yards=None,
            kick_return_fair_catches=None,
            kick_return_fair_catch_percentage=None,
            kick_return_fumbles=None,
            kick_return_fumbles_lost=None,
            kick_returns=None,
            kick_return_touchdowns=None,
            kick_return_yards=None,
            long_kick_return=None,
            long_punt_return=None,
            misc_fumble_returns=None,
            misc_fumble_return_yards=None,
            opposition_fumble_recoveries=None,
            opposition_fumble_recovery_yards=None,
            opposition_special_team_fumble_returns=None,
            opposition_special_team_fumble_return_yards=None,
            punt_return_fair_catches=None,
            punt_return_fair_catch_percentage=None,
            punt_return_fumbles=None,
            punt_return_fumbles_lost=None,
            punt_returns=None,
            punt_returns_started_inside_the_10=None,
            punt_returns_started_inside_the_20=None,
            punt_return_touchdowns=None,
            special_team_fumble_returns=None,
            yards_per_kick_return=None,
            yards_per_punt_return=None,
            yards_per_return=None,
            average_punt_return_yards=None,
            gross_average_punt_yards=None,
            long_punt=None,
            net_average_punt_yards=None,
            punts=None,
            punts_blocked=None,
            punts_blocked_percentage=None,
            punts_inside_10=None,
            punts_inside_10_percentage=None,
            punts_inside_20=None,
            punts_inside_20_percentage=None,
            punts_over_50=None,
            punt_yards=None,
            defensive_points=None,
            misc_points=None,
            return_touchdowns=None,
            total_two_point_conversions=None,
            passing_touchdowns_9_yards=None,
            passing_touchdowns_19_yards=None,
            passing_touchdowns_29_yards=None,
            passing_touchdowns_39_yards=None,
            passing_touchdowns_49_yards=None,
            passing_touchdowns_above_50_yards=None,
            receiving_touchdowns_9_yards=None,
            receiving_touchdowns_19_yards=None,
            receiving_touchdowns_29_yards=None,
            receiving_touchdowns_39_yards=None,
            punt_return_yards=None,
            receiving_touchdowns_49_yards=None,
            receiving_touchdowns_above_50_yards=None,
            rushing_touchdowns_9_yards=None,
            rushing_touchdowns_19_yards=None,
            rushing_touchdowns_29_yards=None,
            rushing_touchdowns_39_yards=None,
            rushing_touchdowns_49_yards=None,
            rushing_touchdowns_above_50_yards=None,
            penalties_in_minutes=None,
            even_strength_goals=None,
            power_play_goals=None,
            short_handed_goals=None,
            game_winning_goals=None,
            even_strength_assists=None,
            power_play_assists=None,
            short_handed_assists=None,
            shots_on_goal=None,
            shooting_percentage=None,
            shifts=None,
            time_on_ice=None,
            decision=None,
            goals_against=None,
            shots_against=None,
            saves=None,
            save_percentage=None,
            shutouts=None,
            individual_corsi_for_events=None,
            on_shot_ice_for_events=None,
            on_shot_ice_against_events=None,
            corsi_for_percentage=None,
            relative_corsi_for_percentage=None,
            offensive_zone_starts=None,
            defensive_zone_starts=None,
            offensive_zone_start_percentage=None,
            hits=None,
            true_shooting_percentage=None,
            at_bats=None,
            runs_scored=None,
            runs_batted_in=None,
            bases_on_balls=None,
            strikeouts=None,
            plate_appearances=None,
            hits_at_bats=None,
            obp=None,
            slg=None,
            ops=None,
            pitches=None,
            strikes=None,
            win_probability_added=None,
            average_leverage_index=None,
            wpa_plus=None,
            wpa_minus=None,
            cwpa=None,
            acli=None,
            re24=None,
            putouts=None,
            innings_pitched=None,
            earned_runs=None,
            home_runs=None,
            era=None,
            batters_faced=None,
            strikes_by_contact=None,
            strikes_swinging=None,
            strikes_looking=None,
            ground_balls=None,
            fly_balls=None,
            line_drives=None,
            inherited_runners=None,
            inherited_scores=None,
            effective_field_goal_percentage=None,
            penalty_kicks_made=None,
            penalty_kicks_attempted=None,
            shots_total=None,
            shots_on_target=None,
            yellow_cards=None,
            red_cards=None,
            touches=None,
            expected_goals=None,
            non_penalty_expected_goals=None,
            expected_assisted_goals=None,
            shot_creating_actions=None,
            goal_creating_actions=None,
            passes_completed=None,
            passes_attempted=None,
            pass_completion=None,
            progressive_passes=None,
            carries=None,
            progressive_carries=None,
            take_ons_attempted=None,
            successful_take_ons=None,
            total_passing_distance=None,
            progressive_passing_distance=None,
            passes_completed_short=None,
            passes_attempted_short=None,
            pass_completion_short=None,
            passes_completed_medium=None,
            passes_attempted_medium=None,
            pass_completion_medium=None,
            passes_completed_long=None,
            passes_attempted_long=None,
            pass_completion_long=None,
            expected_assists=None,
            key_passes=None,
            passes_into_final_third=None,
            passes_into_penalty_area=None,
            crosses_into_penalty_area=None,
            live_ball_passes=None,
            dead_ball_passes=None,
            passes_from_free_kicks=None,
            through_balls=None,
            switches=None,
            crosses=None,
            throw_ins_taken=None,
            corner_kicks=None,
            inswinging_corner_kicks=None,
            outswinging_corner_kicks=None,
            straight_corner_kicks=None,
            passes_offside=None,
            passes_blocked=None,
            tackles_won=None,
            tackles_in_defensive_third=None,
            tackles_in_middle_third=None,
            tackles_in_attacking_third=None,
            dribblers_tackled=None,
            dribbles_challenged=None,
            percent_of_dribblers_tackled=None,
            challenges_lost=None,
            shots_blocked=None,
            tackles_plus_interceptions=None,
            errors=None,
            touches_in_defensive_penalty_area=None,
            touches_in_defensive_third=None,
            touches_in_middle_third=None,
            touches_in_attacking_third=None,
            touches_in_attacking_penalty_area=None,
            live_ball_touches=None,
            successful_take_on_percentage=None,
            times_tackled_during_take_ons=None,
            tackled_during_take_on_percentage=None,
            total_carrying_distance=None,
            progressive_carrying_distance=None,
            carries_into_final_third=None,
            carries_into_penalty_area=None,
            miscontrols=None,
            dispossessed=None,
            passes_received=None,
            progressive_passes_received=None,
            second_yellow_card=None,
            fouls_committed=None,
            fouls_drawn=None,
            offsides=None,
            penalty_kicks_won=None,
            penalty_kicks_conceded=None,
            own_goals=None,
            ball_recoveries=None,
            aerials_won=None,
            aerials_lost=None,
            percentage_of_aerials_won=None,
            shots_on_target_against=None,
            post_shot_expected_goals=None,
            passes_attempted_minus_goal_kicks=None,
            throws_attempted=None,
            percentage_of_passes_that_were_launched=None,
            average_pass_length=None,
            goal_kicks_attempted=None,
            percentage_of_goal_kicks_that_were_launched=None,
            average_goal_kick_length=None,
            crosses_faced=None,
            crosses_stopped=None,
            percentage_crosses_stopped=None,
            defensive_actions_outside_penalty_area=None,
            average_distance_of_defensive_actions=None,
            three_point_attempt_rate=None,
            batting_style=None,
            bowling_style=None,
            playing_roles=None,
            runs=None,
            balls=None,
            fours=None,
            sixes=None,
            strikerate=None,
            fall_of_wicket_order=None,
            fall_of_wicket_num=None,
            fall_of_wicket_runs=None,
            fall_of_wicket_balls=None,
            fall_of_wicket_overs=None,
            fall_of_wicket_over_number=None,
            ball_over_actual=None,
            ball_over_unique=None,
            ball_total_runs=None,
            ball_batsman_runs=None,
            overs=None,
            maidens=None,
            conceded=None,
            wickets=None,
            economy=None,
            runs_per_ball=None,
            dots=None,
            wides=None,
            no_balls=None,
            free_throw_attempt_rate=None,
            offensive_rebound_percentage=None,
            defensive_rebound_percentage=None,
            total_rebound_percentage=None,
            assist_percentage=None,
            steal_percentage=None,
            block_percentage=None,
            turnover_percentage=None,
            usage_percentage=None,
            offensive_rating=None,
            defensive_rating=None,
            box_plus_minus=None,
            ace_percentage=None,
            double_fault_percentage=None,
            first_serves_in=None,
            first_serve_percentage=None,
            second_serve_percentage=None,
            break_points_saved=None,
            return_points_won_percentage=None,
            winners=None,
            winners_fronthand=None,
            winners_backhand=None,
            unforced_errors=None,
            unforced_errors_fronthand=None,
            unforced_errors_backhand=None,
            serve_points=None,
            serves_won=None,
            serves_aces=None,
            serves_unreturned=None,
            serves_forced_error_percentage=None,
            serves_won_in_three_shots_or_less=None,
            serves_wide_percentage=None,
            serves_body_percentage=None,
            serves_t_percentage=None,
            serves_wide_deuce_percentage=None,
            serves_body_deuce_percentage=None,
            serves_t_deuce_percentage=None,
            serves_wide_ad_percentage=None,
            serves_body_ad_percentage=None,
            serves_t_ad_percentage=None,
            serves_net_percentage=None,
            serves_wide_direction_percentage=None,
            shots_deep_percentage=None,
            shots_deep_wide_percentage=None,
            shots_foot_errors_percentage=None,
            shots_unknown_percentage=None,
            points_won_percentage=None,
            goal_accuracy=None,
            tackles_inside_50=None,
            total_possessions=None,
            score_involvements=None,
            centre_clearances=None,
            disposal_efficiency=None,
            stoppage_clearances=None,
            uncontested_marks=None,
            accurate_crosses=None,
            accurate_long_balls=None,
            accurate_passes=None,
            accurate_through_balls=None,
            cross_percentage=None,
            free_kick_goals=None,
            free_kick_percentage=None,
            free_kick_shots=None,
            game_winning_assists=None,
            headed_goals=None,
            inaccurate_crosses=None,
            inaccurate_long_balls=None,
            inaccurate_passes=None,
            inaccurate_through_balls=None,
            left_footed_shots=None,
            long_ball_percentage=None,
            penalty_kick_goals=None,
            penalty_kick_percentage=None,
            penalty_kicks_missed=None,
            possession_percentage=None,
            possession_time=None,
            right_footed_shots=None,
            shoot_out_goals=None,
            shoot_out_misses=None,
            shoot_out_percentage=None,
            shot_assists=None,
            shot_percentage=None,
            shots_headed=None,
            shots_off_target=None,
            shots_on_post=None,
            through_ball_percentage=None,
            long_balls=None,
            total_passes=None,
            average_rating_from_editor=None,
            average_rating_from_user=None,
            did_not_play=None,
            draws=None,
            goal_difference=None,
            losses=None,
            lost_corners=None,
            minutes=None,
            pass_percentage=None,
            starts=None,
            sub_ins=None,
            sub_outs=None,
            suspensions=None,
            time_ended=None,
            time_started=None,
            win_percentage=None,
            wins=None,
            won_corners=None,
            clean_sheet=None,
            crosses_caught=None,
            goals_conceded=None,
            partial_clean_sheet=None,
            penalty_kick_conceded=None,
            penalty_kick_save_percentage=None,
            penalty_kicks_faced=None,
            penalty_kicks_saved=None,
            punches=None,
            shoot_out_kicks_faced=None,
            shoot_out_kicks_saved=None,
            shoot_out_save_percentage=None,
            shots_faced=None,
            smothers=None,
            unclaimed_crosses=None,
            effective_clearances=None,
            effective_tackles=None,
            ineffective_tackles=None,
            tackle_percentage=None,
            appearances=None,
            average_rating_from_correspondent=None,
            average_rating_from_data_feed=None,
            games_played=None,
            team_games_played=None,
            hit_by_pitch=None,
            rbis=None,
            sac_hits=None,
            stolen_bases=None,
            walks=None,
            catcher_interference=None,
            gidps=None,
            sac_flies=None,
            grand_slam_home_runs=None,
            runners_left_on_base=None,
            triples=None,
            game_winning_rbis=None,
            intentional_walks=None,
            doubles=None,
            caught_stealing=None,
            games_started=None,
            pinch_at_bats=None,
            pinch_hits=None,
            player_rating=None,
            is_qualified=None,
            is_qualified_steals=None,
            total_bases=None,
            projected_home_runs=None,
            extra_base_hits=None,
            runs_created=None,
            batting_average=None,
            pinch_average=None,
            slug_average=None,
            secondary_average=None,
            on_base_percentage=None,
            ground_to_fly_ratio=None,
            runs_created_per_27_outs=None,
            batter_rating=None,
            at_bats_per_home_run=None,
            stolen_base_percentage=None,
            pitches_per_plate_appearance=None,
            isolated_power=None,
            walk_to_strikeout_ratio=None,
            walks_per_plate_appearance=None,
            secondary_average_minus_batting_average=None,
            runs_produced=None,
            runs_ratio=None,
            patience_ratio=None,
            balls_in_play_average=None,
            mlb_rating=None,
            offensive_wins_above_replacement=None,
            wins_above_replacement=None,
            batters_hit=None,
            sacrifice_bunts=None,
            save_opportunities=None,
            finishes=None,
            balks=None,
            holds=None,
            complete_games=None,
            perfect_games=None,
            wild_pitches=None,
            third_innings=None,
            team_earned_runs=None,
            pickoff_attempts=None,
            run_support=None,
            pitches_as_starter=None,
            average_game_score=None,
            quality_starts=None,
            inherited_runners_scored=None,
            opponent_total_bases=None,
            is_qualified_saves=None,
            full_innings=None,
            part_innings=None,
            blown_saves=None,
            innings=None,
            whip=None,
            caught_stealing_percentage=None,
            pitches_per_start=None,
            pitches_per_inning=None,
            run_support_average=None,
            opponent_average=None,
            opponent_slug_average=None,
            opponent_on_base_percentage=None,
            opponent_ops=None,
            strikeouts_per_nine_innings=None,
            strikeout_to_walk_ratio=None,
            tough_losses=None,
            cheap_wins=None,
            save_opportunities_per_win=None,
            pitch_count=None,
            strike_pitch_ratio=None,
            double_plays=None,
            opportunities=None,
            passed_balls=None,
            outfield_assists=None,
            pickoffs=None,
            outs_on_field=None,
            triple_plays=None,
            balls_in_zone=None,
            extra_bases=None,
            outs_made=None,
            catcher_third_innings_played=None,
            catcher_caught_stealing=None,
            catcher_stolen_bases_allowed=None,
            catcher_earned_runs=None,
            is_qualified_catcher=None,
            is_qualified_pitcher=None,
            successful_chances=None,
            total_chances=None,
            full_innings_played=None,
            part_innings_played=None,
            fielding_percentage=None,
            range_factor=None,
            zone_rating=None,
            catcher_caught_stealing_percentage=None,
            catcher_era=None,
            def_warbr=None,
            average_three_point_field_goals_attempted=None,
            average_free_throws_made=None,
            average_free_throws_attempted=None,
            average_points=None,
            average_offensive_rebounds=None,
            average_assists=None,
            average_turnovers=None,
            estimated_possessions=None,
            average_estimated_possessions=None,
            points_per_estimated_possessions=None,
            average_team_turnovers=None,
            average_total_turnovers=None,
            three_point_field_goal_percentage=None,
            two_point_field_goals_made=None,
            two_point_field_goals_attempted=None,
            average_two_point_field_goals_made=None,
            average_two_point_field_goals_attempted=None,
            two_point_field_goal_percentage=None,
            shooting_efficiency=None,
            scoring_efficiency=None,
            average_48_field_goals_made=None,
            average_48_field_goals_attempted=None,
            average_48_three_point_field_goals_made=None,
            average_48_three_point_field_goals_attempted=None,
            average_48_free_throws_made=None,
            average_48_free_throws_attempted=None,
            average_48_points=None,
            average_48_offensive_rebounds=None,
            average_48_assists=None,
            average_48_turnovers=None,
            p40=None,
            a40=None,
            average_rebounds=None,
            average_fouls=None,
            average_flagrant_fouls=None,
            average_technical_fouls=None,
            average_ejections=None,
            average_disqualifications=None,
            assist_turnover_ratio=None,
            steal_foul_ratio=None,
            block_foul_ratio=None,
            average_team_rebounds=None,
            total_technical_fouls=None,
            team_assist_turnover_ratio=None,
            steal_turnover_ratio=None,
            average_48_rebounds=None,
            average_48_fouls=None,
            average_48_flagrant_fouls=None,
            average_48_technical_fouls=None,
            average_48_ejections=None,
            average_48_disqualifications=None,
            r40=None,
            double_double=None,
            triple_double=None,
            free_throws_made=None,
            three_point_percentage=None,
            three_point_field_goals_made=None,
            total_turnovers=None,
            points_in_paint=None,
            brick_index=None,
            average_field_goals_made=None,
            average_field_goals_attempted=None,
            average_three_point_field_goals_made=None,
            average_defensive_rebounds=None,
            average_blocks=None,
            average_steals=None,
            average_48_defensive_rebounds=None,
            average_48_blocks=None,
            average_48_steals=None,
            largest_lead=None,
            disqualifications=None,
            flagrant_fouls=None,
            fouls=None,
            ejections=None,
            technical_fouls=None,
            average_minutes=None,
            nba_rating=None,
            plus_minus=None,
            even_strength_time_on_ice=None,
            shifts_per_game=None,
            production=None,
            shot_differential=None,
            goal_differential=None,
            pim_differential=None,
            rating=None,
            average_goals=None,
            ytd_goals=None,
            shots_in_first_period=None,
            shots_in_second_period=None,
            shots_in_third_period=None,
            shots_overtime=None,
            average_shots=None,
            points_per_game=None,
            power_play_opportunities=None,
            power_play_percentage=None,
            shootout_attempts=None,
            shootout_shot_percentage=None,
            empty_net_goals_for=None,
            shutouts_against=None,
            total_face_offs=None,
            average_goals_against=None,
            average_shots_against=None,
            penalty_kill_percentage=None,
            power_play_goals_against=None,
            short_handed_goals_against=None,
            shootout_saves=None,
            shootout_shots_against=None,
            times_short_handed=None,
            empty_net_goals_against=None,
            overtime_losses=None,
            shots_missed=None,
            minor_penalties=None,
            match_penalties=None,
            misconducts=None,
            game_misconducts=None,
            boarding_penalties=None,
            unsportsmanlike_penalties=None,
            fighting_penalties=None,
            average_fights=None,
            time_between_fights=None,
            instigator_penalties=None,
            charging_penalties=None,
            hooking_penalties=None,
            tripping_penalties=None,
            roughing_penalties=None,
            holding_penalties=None,
            interference_penalties=None,
            slashing_penalties=None,
            high_sticking_penalties=None,
            cross_checking_penalties=None,
            stick_holding_penalties=None,
            goalie_interference_penalties=None,
            elbowing_penalties=None,
            diving_penalties=None,
            takeaways=None,
            even_strength_saves=None,
            power_play_saves=None,
            short_handed_saves=None,
            games=None,
            game_started=None,
            ties=None,
            time_on_ice_per_game=None,
            power_play_time_on_ice=None,
            short_handed_time_on_ice=None,
            faceoffs_won=None,
            faceoffs_lost=None,
            faceoff_percentage=None,
            unassisted_goals=None,
            game_tying_goals=None,
            giveaways=None,
            penalties=None,
            penalty_minutes=None,
            penalty_minutes_against=None,
            major_penalties=None,
            net_passing_yards_per_game=None,
            net_yards_per_game=None,
            passing_yards_per_game=None,
            total_points_per_game=None,
            yards_from_scrimmage_per_game=None,
            yards_per_game=None,
            espn_rb_rating=None,
            rushing_yards_per_game=None,
            receiving_yards_per_game=None,
            two_point_returns=None,
            field_goal_attempts=None,
            special_team_fumble_return_yards=None,
            kick_extra_points=None,
            kick_extra_points_made=None,
            attempts_in_box=None,
            second_assists=None,
            qbr=None,
            attempts_out_box=None,
            adjusted_qbr=None,
            turnover_points=None,
            fantasy_rating=None,
            team_turnovers=None,
            second_chance_points=None,
            fast_break_points=None,
            team_rebounds=None,
            gained=None,
        )

    def test_field_goals_attempted(self):
        self.assertEqual(self._player_model.field_goals_attempted, 10)



"""Combined game model."""

# pylint: disable=too-many-locals,line-too-long,too-many-arguments,too-many-branches,too-many-statements,too-many-positional-arguments
import datetime
import logging
from typing import Any

import requests
from pydantic import BaseModel

from ..field_type import TYPE_KEY, FieldType
from ..game_model import VERSION, GameModel
from ..league import League
from ..player_model import PlayerModel
from ..team_model import TeamModel
from ..umpire_model import UmpireModel
from ..venue_model import VenueModel
from .combined_team_model import create_combined_team_model
from .combined_umpire_model import create_combined_umpire_model
from .combined_venue_model import create_combined_venue_model
from .most_interesting import more_interesting
from .normalise_name import normalise_name


def _venue_models(
    game_models: list[GameModel], venue_identity_map: dict[str, str]
) -> tuple[list[VenueModel], str | None]:
    venue_models = []
    full_venue_identity = None
    for game_model in game_models:
        game_model_venue = game_model.venue
        if game_model_venue is not None:
            venue_identity = venue_identity_map.get(game_model_venue.identifier)
            if venue_identity is None:
                logging.warning(
                    "Failed to find %s venue identifier (%s).",
                    game_model_venue.identifier,
                    game_model_venue.name,
                )
            else:
                full_venue_identity = venue_identity
            venue_models.append(game_model_venue)
    return venue_models, full_venue_identity


def _team_models(
    game_models: list[GameModel],
    team_identity_map: dict[str, str],
    player_identity_map: dict[str, str],
    names: dict[str, str],
    coach_names: dict[str, str],
    player_ffill: dict[str, dict[str, Any]],
    team_ffill: dict[str, dict[str, Any]],
    coach_ffill: dict[str, dict[str, Any]],
) -> list[TeamModel]:
    team_models: dict[str, list[TeamModel]] = {}
    for game_model in game_models:
        game_model_teams = game_model.teams
        if game_model_teams:
            for team_model in game_model_teams:
                team_identity = team_identity_map.get(team_model.identifier)
                if team_identity is None:
                    logging.warning(
                        "Failed to find %s team identifier (%s).",
                        team_model.identifier,
                        team_model.name,
                    )
                    team_identity = team_model.identifier
                team_models[team_identity] = team_models.get(team_identity, []) + [
                    team_model
                ]
    return [
        create_combined_team_model(
            v,
            k,
            player_identity_map,
            names,
            coach_names,
            player_ffill,
            team_ffill,
            coach_ffill,
        )  # pyright: ignore
        for k, v in team_models.items()
    ]


def _remove_lookaheads(obj: PlayerModel) -> PlayerModel:
    """Set all fields marked as LOOKAHEAD to None, recursively."""
    if isinstance(obj, BaseModel):
        for name, f in obj.model_fields.items():
            extra = getattr(f, "json_schema_extra", None) or {}
            if extra.get(TYPE_KEY) == FieldType.LOOKAHEAD:
                # validate_assignment=False means this won't revalidate
                setattr(obj, name, None)
            else:
                # Recurse into nested values
                val = getattr(obj, name, None)
                _remove_lookaheads(val)  # type: ignore

    elif isinstance(obj, list):
        for item in obj:  # type: ignore
            _remove_lookaheads(item)

    elif isinstance(obj, dict):
        for v in obj.values():  # type: ignore
            _remove_lookaheads(v)

    return obj


def create_combined_game_model(
    game_models: list[GameModel],
    venue_identity_map: dict[str, str],
    team_identity_map: dict[str, str],
    player_identity_map: dict[str, str],
    session: requests.Session,
    names: dict[str, str],
    coach_names: dict[str, str],
    last_game_number: int | None,
    player_ffill: dict[str, dict[str, Any]],
    team_ffill: dict[str, dict[str, Any]],
    coach_ffill: dict[str, dict[str, Any]],
    umpire_ffill: dict[str, dict[str, Any]],
    team_players_ffill: dict[str, list[PlayerModel]],
) -> GameModel:
    """Create a game model by combining many game models."""
    umpires: dict[str, list[UmpireModel]] = {}
    venue_models, full_venue_identity = _venue_models(game_models, venue_identity_map)
    full_team_models = _team_models(
        game_models,
        team_identity_map,
        player_identity_map,
        names,
        coach_names,
        player_ffill,
        team_ffill,
        coach_ffill,
    )
    attendance = None
    end_dt = None
    year = None
    season_type = None
    week = None
    game_number = None
    postponed = None
    play_off = None
    distance = None
    dividends = []
    pot = None
    best_of = None
    dt_votes: dict[str, int] = {}
    for game_model in game_models:
        dt_votes[game_model.dt.isoformat()] = (
            dt_votes.get(game_model.dt.isoformat(), 0) + 1
        )
        attendance = more_interesting(attendance, game_model.attendance)
        end_dt = more_interesting(end_dt, game_model.end_dt)
        year = more_interesting(year, game_model.year)
        season_type = more_interesting(season_type, game_model.season_type)
        week = more_interesting(week, game_model.week)
        game_number = more_interesting(game_number, game_model.game_number)
        postponed = more_interesting(postponed, game_model.postponed)
        play_off = more_interesting(play_off, game_model.play_off)
        distance = more_interesting(distance, game_model.distance)
        dividends.extend(game_model.dividends)
        pot = more_interesting(pot, game_model.pot)
        best_of = more_interesting(best_of, game_model.best_of)
        for umpire_model in game_model.umpires:
            umpire_id = umpire_model.identifier
            umpire_name_key = normalise_name(umpire_model.name)
            if umpire_name_key in names:
                umpire_id = names[umpire_name_key]
            else:
                names[umpire_name_key] = umpire_id
            umpires[umpire_id] = umpires.get(umpire_id, []) + [umpire_model]

    dt = None
    for dt_iso, _ in sorted(dt_votes.items(), key=lambda x: x[1], reverse=True):
        dt = more_interesting(dt, datetime.datetime.fromisoformat(dt_iso))
    if dt is None:
        raise ValueError("dt is null")

    if full_venue_identity is None and venue_models:
        for venue_model in venue_models:
            venue_model_identifier = venue_model.identifier
            if venue_model_identifier is not None:
                full_venue_identity = venue_model_identifier

    if game_number is None and last_game_number is not None:
        game_number = last_game_number + 1

    league = game_models[0].league

    # Forward fill for team player models
    if league == str(League.NFL):
        for team_model in full_team_models:
            if team_model.players:
                team_players_ffill[team_model.identifier] = team_model.players
                continue
            team_model.players = [
                _remove_lookaheads(x)
                for x in team_players_ffill.get(team_model.identifier, [])
            ]

    return GameModel.model_construct(
        dt=dt,
        week=week,
        game_number=game_number,
        venue=create_combined_venue_model(venue_models, full_venue_identity, session),  # pyright: ignore
        teams=full_team_models,
        end_dt=end_dt,
        attendance=attendance,
        league=league,
        year=year,
        season_type=season_type,
        postponed=postponed,
        play_off=play_off,
        distance=distance,
        dividends=dividends,
        pot=pot,
        version=VERSION,
        umpires=[
            create_combined_umpire_model(v, k, umpire_ffill) for k, v in umpires.items()
        ],
    )



"""NFL NFL.com game model."""

import datetime
from urllib.parse import urlparse

from bs4 import BeautifulSoup
from playwright.sync_api import Playwright
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ....playwright import ensure_install
from ...game_model import GameModel
from ...league import League
from ...team_model import VERSION
from ...venue_model import VERSION as VENUE_VERSION
from .nfl_nflcom_team_model import create_nfl_nflcom_team_model
from .nfl_nflcom_venue_model import create_nfl_nflcom_venue_model


def create_nfl_nflcom_game_model(
    url: str,
    session: ScrapeSession,
    playwright: Playwright,
    version: str,
) -> GameModel:
    """Create a game model from NFL.com."""
    o = urlparse(url)
    end_path_split = o.path.split("/")[-1].split("-")
    week = int(end_path_split[-1])

    ensure_install()
    browser = playwright.chromium.launch()
    context = browser.new_context()
    page = context.new_page()
    page.goto(url, wait_until="load")
    soup = BeautifulSoup(page.content(), "lxml")

    dt = None
    for time in soup.find_all("time"):
        dt = datetime.datetime.fromisoformat(time.get("datetime"))
        break
    if dt is None:
        raise ValueError("dt is null")

    venue_model = None
    for dd in soup.find_all("dd", {"aria-labelledby": "info-card-term-STADIUM"}):
        venue_model = create_nfl_nflcom_venue_model(
            venue_name=dd.get_text().strip(),
            session=session,
            dt=dt,
            version=VENUE_VERSION,
        )
        break

    home_team_name = end_path_split[2]
    away_team_name = end_path_split[0]
    teams = [
        create_nfl_nflcom_team_model(
            team_name=x, dt=dt, session=session, version=VERSION
        )
        for x in [home_team_name, away_team_name]
    ]

    return GameModel(
        dt=dt,
        week=week,
        game_number=None,
        venue=venue_model,
        teams=teams,
        end_dt=None,
        attendance=None,
        league=League.NFL,
        year=dt.year,
        season_type=None,
        postponed=None,
        play_off=None,
        distance=None,
        dividends=[],
        pot=None,
        version=version,
        umpires=[],
        best_of=None,
    )



"""NCAABW combined league model."""

from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.ncaabw_espn_league_model import NCAABWESPNLeagueModel
from ..oddsportal.ncaabw_oddsportal_league_model import \
    NCAABWOddsPortalLeagueModel

NCAABW_TEAM_IDENTITY_MAP: dict[str, str] = {}
NCAABW_VENUE_IDENTITY_MAP: dict[str, str] = {}
NCAABW_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class NCAABWCombinedLeagueModel(CombinedLeagueModel):
    """NCAABW combined implementation of the league model."""

    def __init__(self, session: ScrapeSession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.NCAABW,
            [
                NCAABWESPNLeagueModel(session, position=0),
                NCAABWOddsPortalLeagueModel(session, position=1),
            ],
            league_filter,
        )

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return NCAABW_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return NCAABW_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return NCAABW_PLAYER_IDENTITY_MAP

    @classmethod
    def name(cls) -> str:
        return "ncaabw-combined-league-model"



"""Tests for the combined player model class."""
import os
import unittest

import requests_mock
import requests_cache
from sportsball.data.combined.combined_player_model import create_combined_player_model
from sportsball.data.player_model import PlayerModel, VERSION
from sportsball.data.species import Species


class TestCombinedPlayerModel(unittest.TestCase):

    def setUp(self):
        self._session = requests_cache.CachedSession(backend="memory")
        self.dir = os.path.dirname(__file__)

    def test_field_goals_attempted(self):
        identifier = "a"
        field_goals_attempted = 10
        with requests_mock.Mocker() as m:
            player_model = PlayerModel(
                identifier="a",
                jersey="35",
                kicks=None,
                fumbles=None,
                fumbles_lost=None,
                field_goals=None,
                field_goals_attempted=field_goals_attempted,
                offensive_rebounds=None,
                assists=None,
                turnovers=None,
                name="James Jones",
                marks=None,
                handballs=None,
                disposals=None,
                goals=None,
                behinds=None,
                hit_outs=None,
                tackles=None,
                rebounds=None,
                insides=None,
                clearances=None,
                clangers=None,
                free_kicks_for=None,
                free_kicks_against=None,
                brownlow_votes=None,
                contested_possessions=None,
                uncontested_possessions=None,
                contested_marks=None,
                marks_inside=None,
                one_percenters=None,
                bounces=None,
                goal_assists=None,
                percentage_played=None,
                birth_date=None,
                age=None,
                species=str(Species.HUMAN),
                handicap_weight=None,
                father=None,
                starting_position=None,
                weight=None,
                birth_address=None,
                owner=None,
                seconds_played=None,
                three_point_field_goals=None,
                three_point_field_goals_attempted=None,
                free_throws=None,
                free_throws_attempted=None,
                defensive_rebounds=None,
                steals=None,
                blocks=None,
                personal_fouls=None,
                points=None,
                game_score=None,
                point_differential=None,
                version=VERSION,
                height=None,
                colleges=[],
                headshot=None,
                forced_fumbles=None,
                fumbles_recovered=None,
                fumbles_recovered_yards=None,
                fumbles_touchdowns=None,
                offensive_two_point_returns=None,
                offensive_fumbles_touchdowns=None,
                defensive_fumbles_touchdowns=None,
                average_gain=None,
                completion_percentage=None,
                completions=None,
                espn_quarterback_rating=None,
                interception_percentage=None,
                interceptions=None,
                long_passing=None,
                misc_yards=None,
                net_passing_yards=None,
                net_total_yards=None,
                passing_attempts=None,
                passing_big_plays=None,
                passing_first_downs=None,
                passing_fumbles=None,
                passing_fumbles_lost=None,
                passing_touchdown_percentage=None,
                passing_touchdowns=None,
                passing_yards=None,
                passing_yards_after_catch=None,
                quarterback_rating=None,
                sacks=None,
                passing_yards_at_catch=None,
                sacks_yards_lost=None,
                net_passing_attempts=None,
                total_offensive_plays=None,
                total_points=None,
                total_touchdowns=None,
                total_yards=None,
                total_yards_from_scrimmage=None,
                two_point_pass=None,
                two_point_pass_attempt=None,
                yards_per_completion=None,
                yards_per_pass_attempt=None,
                net_yards_per_pass_attempt=None,
                long_rushing=None,
                rushing_attempts=None,
                rushing_big_plays=None,
                rushing_first_downs=None,
                rushing_fumbles=None,
                rushing_fumbles_lost=None,
                rushing_touchdowns=None,
                rushing_yards=None,
                stuffs=None,
                stuff_yards_lost=None,
                two_point_rush=None,
                two_point_rush_attempts=None,
                yards_per_rush_attempt=None,
                espn_widereceiver=None,
                long_reception=None,
                receiving_big_plays=None,
                receiving_first_downs=None,
                receiving_fumbles=None,
                receiving_fumbles_lost=None,
                receiving_targets=None,
                receiving_touchdowns=None,
                receiving_yards=None,
                receiving_yards_after_catch=None,
                receiving_yards_at_catch=None,
                receptions=None,
                two_point_receptions=None,
                two_point_reception_attempts=None,
                yards_per_reception=None,
                assist_tackles=None,
                average_interception_yards=None,
                average_sack_yards=None,
                average_stuff_yards=None,
                blocked_field_goal_touchdowns=None,
                blocked_punt_touchdowns=None,
                defensive_touchdowns=None,
                hurries=None,
                kicks_blocked=None,
                long_interception=None,
                misc_touchdowns=None,
                passes_batted_down=None,
                passes_defended=None,
                quarterback_hits=None,
                sacks_assisted=None,
                sacks_unassisted=None,
                sacks_yards=None,
                safeties=None,
                solo_tackles=None,
                stuff_yards=None,
                tackles_for_loss=None,
                tackles_yards_lost=None,
                yards_allowed=None,
                points_allowed=None,
                one_point_safeties_made=None,
                missed_field_goal_return_td=None,
                blocked_punt_ez_rec_td=None,
                interception_touchdowns=None,
                interception_yards=None,
                average_kickoff_return_yards=None,
                average_kickoff_yards=None,
                extra_point_attempts=None,
                extra_point_percentage=None,
                extra_point_blocked=None,
                extra_points_blocked_percentage=None,
                extra_points_made=None,
                fair_catches=None,
                fair_catch_percentage=None,
                field_goal_attempts_max_19_yards=None,
                field_goal_attempts_max_29_yards=None,
                field_goal_attempts_max_39_yards=None,
                field_goal_attempts_max_49_yards=None,
                field_goal_attempts_max_59_yards=None,
                field_goal_attempts_max_99_yards=None,
                field_goal_attempts_above_50_yards=None,
                field_goal_attempt_yards=None,
                field_goals_blocked=None,
                field_goals_blocked_percentage=None,
                field_goals_made=None,
                field_goals_made_max_19_yards=None,
                field_goals_made_max_29_yards=None,
                field_goals_made_max_39_yards=None,
                field_goals_made_max_49_yards=None,
                field_goals_made_max_59_yards=None,
                field_goals_made_max_99_yards=None,
                field_goals_made_above_50_yards=None,
                field_goals_made_yards=None,
                field_goals_missed_yards=None,
                kickoff_out_of_bounds=None,
                kickoff_returns=None,
                kickoff_returns_touchdowns=None,
                kickoff_return_yards=None,
                kickoffs=None,
                kickoff_yards=None,
                long_field_goal_attempt=None,
                long_field_goal_made=None,
                long_kickoff=None,
                total_kicking_points=None,
                touchback_percentage=None,
                touchbacks=None,
                defensive_fumble_returns=None,
                defensive_fumble_return_yards=None,
                fumble_recoveries=None,
                fumble_recovery_yards=None,
                kick_return_fair_catches=None,
                kick_return_fair_catch_percentage=None,
                kick_return_fumbles=None,
                kick_return_fumbles_lost=None,
                kick_returns=None,
                kick_return_touchdowns=None,
                kick_return_yards=None,
                long_kick_return=None,
                long_punt_return=None,
                misc_fumble_returns=None,
                misc_fumble_return_yards=None,
                opposition_fumble_recoveries=None,
                opposition_fumble_recovery_yards=None,
                opposition_special_team_fumble_returns=None,
                opposition_special_team_fumble_return_yards=None,
                punt_return_fair_catches=None,
                punt_return_fair_catch_percentage=None,
                punt_return_fumbles=None,
                punt_return_fumbles_lost=None,
                punt_returns=None,
                punt_returns_started_inside_the_10=None,
                punt_returns_started_inside_the_20=None,
                punt_return_touchdowns=None,
                special_team_fumble_returns=None,
                yards_per_kick_return=None,
                yards_per_punt_return=None,
                yards_per_return=None,
                average_punt_return_yards=None,
                gross_average_punt_yards=None,
                long_punt=None,
                net_average_punt_yards=None,
                punts=None,
                punts_blocked=None,
                punts_blocked_percentage=None,
                punts_inside_10=None,
                punts_inside_10_percentage=None,
                punts_inside_20=None,
                punts_inside_20_percentage=None,
                punts_over_50=None,
                punt_yards=None,
                defensive_points=None,
                misc_points=None,
                return_touchdowns=None,
                total_two_point_conversions=None,
                passing_touchdowns_9_yards=None,
                passing_touchdowns_19_yards=None,
                passing_touchdowns_29_yards=None,
                passing_touchdowns_39_yards=None,
                passing_touchdowns_49_yards=None,
                passing_touchdowns_above_50_yards=None,
                receiving_touchdowns_9_yards=None,
                receiving_touchdowns_19_yards=None,
                receiving_touchdowns_29_yards=None,
                receiving_touchdowns_39_yards=None,
                punt_return_yards=None,
                receiving_touchdowns_49_yards=None,
                receiving_touchdowns_above_50_yards=None,
                rushing_touchdowns_9_yards=None,
                rushing_touchdowns_19_yards=None,
                rushing_touchdowns_29_yards=None,
                rushing_touchdowns_39_yards=None,
                rushing_touchdowns_49_yards=None,
                rushing_touchdowns_above_50_yards=None,
                penalties_in_minutes=None,
                even_strength_goals=None,
                power_play_goals=None,
                short_handed_goals=None,
                game_winning_goals=None,
                even_strength_assists=None,
                power_play_assists=None,
                short_handed_assists=None,
                shots_on_goal=None,
                shooting_percentage=None,
                shifts=None,
                time_on_ice=None,
                decision=None,
                goals_against=None,
                shots_against=None,
                saves=None,
                save_percentage=None,
                shutouts=None,
                individual_corsi_for_events=None,
                on_shot_ice_for_events=None,
                on_shot_ice_against_events=None,
                corsi_for_percentage=None,
                relative_corsi_for_percentage=None,
                offensive_zone_starts=None,
                defensive_zone_starts=None,
                offensive_zone_start_percentage=None,
                hits=None,
                true_shooting_percentage=None,
                at_bats=None,
                runs_scored=None,
                runs_batted_in=None,
                bases_on_balls=None,
                strikeouts=None,
                plate_appearances=None,
                hits_at_bats=None,
                obp=None,
                slg=None,
                ops=None,
                pitches=None,
                strikes=None,
                win_probability_added=None,
                average_leverage_index=None,
                wpa_plus=None,
                wpa_minus=None,
                cwpa=None,
                acli=None,
                re24=None,
                putouts=None,
                innings_pitched=None,
                earned_runs=None,
                home_runs=None,
                era=None,
                batters_faced=None,
                strikes_by_contact=None,
                strikes_swinging=None,
                strikes_looking=None,
                ground_balls=None,
                fly_balls=None,
                line_drives=None,
                inherited_runners=None,
                inherited_scores=None,
                effective_field_goal_percentage=None,
                penalty_kicks_made=None,
                penalty_kicks_attempted=None,
                shots_total=None,
                shots_on_target=None,
                yellow_cards=None,
                red_cards=None,
                touches=None,
                expected_goals=None,
                non_penalty_expected_goals=None,
                expected_assisted_goals=None,
                shot_creating_actions=None,
                goal_creating_actions=None,
                passes_completed=None,
                passes_attempted=None,
                pass_completion=None,
                progressive_passes=None,
                carries=None,
                progressive_carries=None,
                take_ons_attempted=None,
                successful_take_ons=None,
                total_passing_distance=None,
                progressive_passing_distance=None,
                passes_completed_short=None,
                passes_attempted_short=None,
                pass_completion_short=None,
                passes_completed_medium=None,
                passes_attempted_medium=None,
                pass_completion_medium=None,
                passes_completed_long=None,
                passes_attempted_long=None,
                pass_completion_long=None,
                expected_assists=None,
                key_passes=None,
                passes_into_final_third=None,
                passes_into_penalty_area=None,
                crosses_into_penalty_area=None,
                live_ball_passes=None,
                dead_ball_passes=None,
                passes_from_free_kicks=None,
                through_balls=None,
                switches=None,
                crosses=None,
                throw_ins_taken=None,
                corner_kicks=None,
                inswinging_corner_kicks=None,
                outswinging_corner_kicks=None,
                straight_corner_kicks=None,
                passes_offside=None,
                passes_blocked=None,
                tackles_won=None,
                tackles_in_defensive_third=None,
                tackles_in_middle_third=None,
                tackles_in_attacking_third=None,
                dribblers_tackled=None,
                dribbles_challenged=None,
                percent_of_dribblers_tackled=None,
                challenges_lost=None,
                shots_blocked=None,
                tackles_plus_interceptions=None,
                errors=None,
                touches_in_defensive_penalty_area=None,
                touches_in_defensive_third=None,
                touches_in_middle_third=None,
                touches_in_attacking_third=None,
                touches_in_attacking_penalty_area=None,
                live_ball_touches=None,
                successful_take_on_percentage=None,
                times_tackled_during_take_ons=None,
                tackled_during_take_on_percentage=None,
                total_carrying_distance=None,
                progressive_carrying_distance=None,
                carries_into_final_third=None,
                carries_into_penalty_area=None,
                miscontrols=None,
                dispossessed=None,
                passes_received=None,
                progressive_passes_received=None,
                second_yellow_card=None,
                fouls_committed=None,
                fouls_drawn=None,
                offsides=None,
                penalty_kicks_won=None,
                penalty_kicks_conceded=None,
                own_goals=None,
                ball_recoveries=None,
                aerials_won=None,
                aerials_lost=None,
                percentage_of_aerials_won=None,
                shots_on_target_against=None,
                post_shot_expected_goals=None,
                passes_attempted_minus_goal_kicks=None,
                throws_attempted=None,
                percentage_of_passes_that_were_launched=None,
                average_pass_length=None,
                goal_kicks_attempted=None,
                percentage_of_goal_kicks_that_were_launched=None,
                average_goal_kick_length=None,
                crosses_faced=None,
                crosses_stopped=None,
                percentage_crosses_stopped=None,
                defensive_actions_outside_penalty_area=None,
                average_distance_of_defensive_actions=None,
                three_point_attempt_rate=None,
                batting_style=None,
                bowling_style=None,
                playing_roles=None,
                runs=None,
                balls=None,
                fours=None,
                sixes=None,
                strikerate=None,
                fall_of_wicket_order=None,
                fall_of_wicket_num=None,
                fall_of_wicket_runs=None,
                fall_of_wicket_balls=None,
                fall_of_wicket_overs=None,
                fall_of_wicket_over_number=None,
                ball_over_actual=None,
                ball_over_unique=None,
                ball_total_runs=None,
                ball_batsman_runs=None,
                overs=None,
                maidens=None,
                conceded=None,
                wickets=None,
                economy=None,
                runs_per_ball=None,
                dots=None,
                wides=None,
                no_balls=None,
                free_throw_attempt_rate=None,
                offensive_rebound_percentage=None,
                defensive_rebound_percentage=None,
                total_rebound_percentage=None,
                assist_percentage=None,
                steal_percentage=None,
                block_percentage=None,
                turnover_percentage=None,
                usage_percentage=None,
                offensive_rating=None,
                defensive_rating=None,
                box_plus_minus=None,
                ace_percentage=None,
                double_fault_percentage=None,
                first_serves_in=None,
                first_serve_percentage=None,
                second_serve_percentage=None,
                break_points_saved=None,
                return_points_won_percentage=None,
                winners=None,
                winners_fronthand=None,
                winners_backhand=None,
                unforced_errors=None,
                unforced_errors_fronthand=None,
                unforced_errors_backhand=None,
                serve_points=None,
                serves_won=None,
                serves_aces=None,
                serves_unreturned=None,
                serves_forced_error_percentage=None,
                serves_won_in_three_shots_or_less=None,
                serves_wide_percentage=None,
                serves_body_percentage=None,
                serves_t_percentage=None,
                serves_wide_deuce_percentage=None,
                serves_body_deuce_percentage=None,
                serves_t_deuce_percentage=None,
                serves_wide_ad_percentage=None,
                serves_body_ad_percentage=None,
                serves_t_ad_percentage=None,
                serves_net_percentage=None,
                serves_wide_direction_percentage=None,
                shots_deep_percentage=None,
                shots_deep_wide_percentage=None,
                shots_foot_errors_percentage=None,
                shots_unknown_percentage=None,
                points_won_percentage=None,
                goal_accuracy=None,
                tackles_inside_50=None,
                total_possessions=None,
                score_involvements=None,
                centre_clearances=None,
                disposal_efficiency=None,
                stoppage_clearances=None,
                uncontested_marks=None,
                accurate_crosses=None,
                accurate_long_balls=None,
                accurate_passes=None,
                accurate_through_balls=None,
                cross_percentage=None,
                free_kick_goals=None,
                free_kick_percentage=None,
                free_kick_shots=None,
                game_winning_assists=None,
                headed_goals=None,
                inaccurate_crosses=None,
                inaccurate_long_balls=None,
                inaccurate_passes=None,
                inaccurate_through_balls=None,
                left_footed_shots=None,
                long_ball_percentage=None,
                penalty_kick_goals=None,
                penalty_kick_percentage=None,
                penalty_kicks_missed=None,
                possession_percentage=None,
                possession_time=None,
                right_footed_shots=None,
                shoot_out_goals=None,
                shoot_out_misses=None,
                shoot_out_percentage=None,
                shot_assists=None,
                shot_percentage=None,
                shots_headed=None,
                shots_off_target=None,
                shots_on_post=None,
                through_ball_percentage=None,
                long_balls=None,
                total_passes=None,
                average_rating_from_editor=None,
                average_rating_from_user=None,
                did_not_play=None,
                draws=None,
                goal_difference=None,
                losses=None,
                lost_corners=None,
                minutes=None,
                pass_percentage=None,
                starts=None,
                sub_ins=None,
                sub_outs=None,
                suspensions=None,
                time_ended=None,
                time_started=None,
                win_percentage=None,
                wins=None,
                won_corners=None,
                clean_sheet=None,
                crosses_caught=None,
                goals_conceded=None,
                partial_clean_sheet=None,
                penalty_kick_conceded=None,
                penalty_kick_save_percentage=None,
                penalty_kicks_faced=None,
                penalty_kicks_saved=None,
                punches=None,
                shoot_out_kicks_faced=None,
                shoot_out_kicks_saved=None,
                shoot_out_save_percentage=None,
                shots_faced=None,
                smothers=None,
                unclaimed_crosses=None,
                effective_clearances=None,
                effective_tackles=None,
                ineffective_tackles=None,
                tackle_percentage=None,
                appearances=None,
                average_rating_from_correspondent=None,
                average_rating_from_data_feed=None,
                games_played=None,
                team_games_played=None,
                hit_by_pitch=None,
                rbis=None,
                sac_hits=None,
                stolen_bases=None,
                walks=None,
                catcher_interference=None,
                gidps=None,
                sac_flies=None,
                grand_slam_home_runs=None,
                runners_left_on_base=None,
                triples=None,
                game_winning_rbis=None,
                intentional_walks=None,
                doubles=None,
                caught_stealing=None,
                games_started=None,
                pinch_at_bats=None,
                pinch_hits=None,
                player_rating=None,
                is_qualified=None,
                is_qualified_steals=None,
                total_bases=None,
                projected_home_runs=None,
                extra_base_hits=None,
                runs_created=None,
                batting_average=None,
                pinch_average=None,
                slug_average=None,
                secondary_average=None,
                on_base_percentage=None,
                ground_to_fly_ratio=None,
                runs_created_per_27_outs=None,
                batter_rating=None,
                at_bats_per_home_run=None,
                stolen_base_percentage=None,
                pitches_per_plate_appearance=None,
                isolated_power=None,
                walk_to_strikeout_ratio=None,
                walks_per_plate_appearance=None,
                secondary_average_minus_batting_average=None,
                runs_produced=None,
                runs_ratio=None,
                patience_ratio=None,
                balls_in_play_average=None,
                mlb_rating=None,
                offensive_wins_above_replacement=None,
                wins_above_replacement=None,
                batters_hit=None,
                sacrifice_bunts=None,
                save_opportunities=None,
                finishes=None,
                balks=None,
                holds=None,
                complete_games=None,
                perfect_games=None,
                wild_pitches=None,
                third_innings=None,
                team_earned_runs=None,
                pickoff_attempts=None,
                run_support=None,
                pitches_as_starter=None,
                average_game_score=None,
                quality_starts=None,
                inherited_runners_scored=None,
                opponent_total_bases=None,
                is_qualified_saves=None,
                full_innings=None,
                part_innings=None,
                blown_saves=None,
                innings=None,
                whip=None,
                caught_stealing_percentage=None,
                pitches_per_start=None,
                pitches_per_inning=None,
                run_support_average=None,
                opponent_average=None,
                opponent_slug_average=None,
                opponent_on_base_percentage=None,
                opponent_ops=None,
                strikeouts_per_nine_innings=None,
                strikeout_to_walk_ratio=None,
                tough_losses=None,
                cheap_wins=None,
                save_opportunities_per_win=None,
                pitch_count=None,
                strike_pitch_ratio=None,
                double_plays=None,
                opportunities=None,
                passed_balls=None,
                outfield_assists=None,
                pickoffs=None,
                outs_on_field=None,
                triple_plays=None,
                balls_in_zone=None,
                extra_bases=None,
                outs_made=None,
                catcher_third_innings_played=None,
                catcher_caught_stealing=None,
                catcher_stolen_bases_allowed=None,
                catcher_earned_runs=None,
                is_qualified_catcher=None,
                is_qualified_pitcher=None,
                successful_chances=None,
                total_chances=None,
                full_innings_played=None,
                part_innings_played=None,
                fielding_percentage=None,
                range_factor=None,
                zone_rating=None,
                catcher_caught_stealing_percentage=None,
                catcher_era=None,
                def_warbr=None,
                average_three_point_field_goals_attempted=None,
                average_free_throws_made=None,
                average_free_throws_attempted=None,
                average_points=None,
                average_offensive_rebounds=None,
                average_assists=None,
                average_turnovers=None,
                estimated_possessions=None,
                average_estimated_possessions=None,
                points_per_estimated_possessions=None,
                average_team_turnovers=None,
                average_total_turnovers=None,
                three_point_field_goal_percentage=None,
                two_point_field_goals_made=None,
                two_point_field_goals_attempted=None,
                average_two_point_field_goals_made=None,
                average_two_point_field_goals_attempted=None,
                two_point_field_goal_percentage=None,
                shooting_efficiency=None,
                scoring_efficiency=None,
                average_48_field_goals_made=None,
                average_48_field_goals_attempted=None,
                average_48_three_point_field_goals_made=None,
                average_48_three_point_field_goals_attempted=None,
                average_48_free_throws_made=None,
                average_48_free_throws_attempted=None,
                average_48_points=None,
                average_48_offensive_rebounds=None,
                average_48_assists=None,
                average_48_turnovers=None,
                p40=None,
                a40=None,
                average_rebounds=None,
                average_fouls=None,
                average_flagrant_fouls=None,
                average_technical_fouls=None,
                average_ejections=None,
                average_disqualifications=None,
                assist_turnover_ratio=None,
                steal_foul_ratio=None,
                block_foul_ratio=None,
                average_team_rebounds=None,
                total_technical_fouls=None,
                team_assist_turnover_ratio=None,
                steal_turnover_ratio=None,
                average_48_rebounds=None,
                average_48_fouls=None,
                average_48_flagrant_fouls=None,
                average_48_technical_fouls=None,
                average_48_ejections=None,
                average_48_disqualifications=None,
                r40=None,
                double_double=None,
                triple_double=None,
                free_throws_made=None,
                three_point_percentage=None,
                three_point_field_goals_made=None,
                total_turnovers=None,
                points_in_paint=None,
                brick_index=None,
                average_field_goals_made=None,
                average_field_goals_attempted=None,
                average_three_point_field_goals_made=None,
                average_defensive_rebounds=None,
                average_blocks=None,
                average_steals=None,
                average_48_defensive_rebounds=None,
                average_48_blocks=None,
                average_48_steals=None,
                largest_lead=None,
                disqualifications=None,
                flagrant_fouls=None,
                fouls=None,
                ejections=None,
                technical_fouls=None,
                average_minutes=None,
                nba_rating=None,
                plus_minus=None,
                even_strength_time_on_ice=None,
                shifts_per_game=None,
                production=None,
                shot_differential=None,
                goal_differential=None,
                pim_differential=None,
                rating=None,
                average_goals=None,
                ytd_goals=None,
                shots_in_first_period=None,
                shots_in_second_period=None,
                shots_in_third_period=None,
                shots_overtime=None,
                average_shots=None,
                points_per_game=None,
                power_play_opportunities=None,
                power_play_percentage=None,
                shootout_attempts=None,
                shootout_shot_percentage=None,
                empty_net_goals_for=None,
                shutouts_against=None,
                total_face_offs=None,
                average_goals_against=None,
                average_shots_against=None,
                penalty_kill_percentage=None,
                power_play_goals_against=None,
                short_handed_goals_against=None,
                shootout_saves=None,
                shootout_shots_against=None,
                times_short_handed=None,
                empty_net_goals_against=None,
                overtime_losses=None,
                shots_missed=None,
                minor_penalties=None,
                match_penalties=None,
                misconducts=None,
                game_misconducts=None,
                boarding_penalties=None,
                unsportsmanlike_penalties=None,
                fighting_penalties=None,
                average_fights=None,
                time_between_fights=None,
                instigator_penalties=None,
                charging_penalties=None,
                hooking_penalties=None,
                tripping_penalties=None,
                roughing_penalties=None,
                holding_penalties=None,
                interference_penalties=None,
                slashing_penalties=None,
                high_sticking_penalties=None,
                cross_checking_penalties=None,
                stick_holding_penalties=None,
                goalie_interference_penalties=None,
                elbowing_penalties=None,
                diving_penalties=None,
                takeaways=None,
                even_strength_saves=None,
                power_play_saves=None,
                short_handed_saves=None,
                games=None,
                game_started=None,
                ties=None,
                time_on_ice_per_game=None,
                power_play_time_on_ice=None,
                short_handed_time_on_ice=None,
                faceoffs_won=None,
                faceoffs_lost=None,
                faceoff_percentage=None,
                unassisted_goals=None,
                game_tying_goals=None,
                giveaways=None,
                penalties=None,
                penalty_minutes=None,
                penalty_minutes_against=None,
                major_penalties=None,
                net_passing_yards_per_game=None,
                net_yards_per_game=None,
                passing_yards_per_game=None,
                total_points_per_game=None,
                yards_from_scrimmage_per_game=None,
                yards_per_game=None,
                espn_rb_rating=None,
                rushing_yards_per_game=None,
                receiving_yards_per_game=None,
                two_point_returns=None,
                field_goal_attempts=None,
                special_team_fumble_return_yards=None,
                kick_extra_points=None,
                kick_extra_points_made=None,
                attempts_in_box=None,
                second_assists=None,
                qbr=None,
                attempts_out_box=None,
                adjusted_qbr=None,
                turnover_points=None,
                fantasy_rating=None,
                second_chance_points=None,
                team_turnovers=None,
                fast_break_points=None,
                team_rebounds=None,
                gained=None,
            )
            players_ffill = {}
            combined_player_model = create_combined_player_model(
                player_models=[player_model],
                identifier=identifier,
                player_ffill=players_ffill,
            )
            self.assertEqual(combined_player_model.field_goals_attempted, field_goals_attempted)



"""NCAAF sports reference league model."""

# pylint: disable=line-too-long

from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...league import League
from ...nfl.position import Position
from ...sportsreference.sportsreference_league_model import \
    SportsReferenceLeagueModel


class NCAAFSportsReferenceLeagueModel(SportsReferenceLeagueModel):
    """NCAAF Sports Reference implementation of the league model."""

    def __init__(self, session: ScrapeSession, position: int | None = None) -> None:
        super().__init__(
            session,
            League.NCAAF,
            "https://www.sports-reference.com/cfb/boxscores/",
            position=position,
        )

    @classmethod
    def name(cls) -> str:
        return "ncaaf-sportsreference-league-model"

    @classmethod
    def position_validator(cls) -> dict[str, str]:
        return {str(x): str(x) for x in Position}



"""HKJC HKJC game model."""

# pylint: disable=too-many-nested-blocks,too-many-statements,too-many-branches,too-many-locals,too-many-return-statements,too-many-boolean-expressions,use-maxsplit-arg
import datetime
import io
import logging
import urllib.parse
from urllib.parse import urlparse

import pandas as pd
import pytest_is_running
from bs4 import BeautifulSoup
from dateutil.parser import parse
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ....cache import MEMORY
from ...game_model import VERSION, GameModel
from ...league import League
from ...venue_model import VERSION as VENUE_VERSION
from ..position import position_from_str
from .hkjc_hkjc_dividend_model import create_hkjc_hkjc_dividend_model
from .hkjc_hkjc_odds_model import create_hkjc_hkjc_odds_model
from .hkjc_hkjc_team_model import create_hkjc_hkjc_team_model
from .hkjc_hkjc_venue_model import create_hkjc_hkjc_venue_model

RACE_DATE_QUERY_KEY = "RaceDate"
RACE_NUMBER_QUERY_KEY = "RaceNo"
RACE_COURSE_QUERY_KEY = "Racecourse"


def _create_hkjc_hkjc_game_model(
    session: ScrapeSession,
    html: str,
    url: str,
    version: str,
) -> GameModel | None:
    soup = BeautifulSoup(html, "lxml")
    for div in soup.find_all("div", {"id": "errorContainer"}):
        div_text = div.get_text().strip()
        if div_text.lower() == "no information.":
            return None
    for div in soup.find_all("div"):
        div_text = div.get_text().strip()
        if "this race is declared abandoned" in div_text.lower():
            return None
        if "this race is declared void" in div_text.lower():
            return None
        if "information will be released shortly" in div_text.lower():
            return None
        if "unofficial dividends" in div_text.lower():
            return None
        if "this race has been abandoned" in div_text.lower():
            return None

    o = urlparse(url)
    query = urllib.parse.parse_qs(o.query)
    try:
        dt = parse(query[RACE_DATE_QUERY_KEY][0])
    except KeyError as exc:
        logging.warning(url)
        logging.warning(str(exc))
        return None
    game_number = int(query[RACE_NUMBER_QUERY_KEY][0])
    venue_code = query[RACE_COURSE_QUERY_KEY][0]

    handle = io.StringIO()
    handle.write(html)
    handle.seek(0)
    dfs = pd.read_html(handle)
    race_track = None
    distance = None
    pot = None
    teams = []
    dividends = []
    for count, df in enumerate(dfs):
        if count == 1:
            race_track = ""
            try:
                race_track = str(df.iat[2, 2]).split("-", maxsplit=1)[0].strip()
            except IndexError:
                logging.error(url)
                raise

            distance = 0.0
            try:
                distance = float(df.iat[1, 0].split("-")[1].strip().replace("M", ""))
            except IndexError:
                logging.error(html)
                logging.error(url)
                raise
            pot = float(df.iat[3, 0].split()[-1].strip().replace(",", ""))
        elif count == 2:
            for _, row in df.iterrows():
                horse_name = None
                try:
                    horse_name = "(".join(row["Horse"].split("(")[:-1]).strip()
                except KeyError:
                    logging.error(url)
                    continue

                jockey_name = row["Jockey"].split("(")[0].strip()

                trainer_name = None
                try:
                    trainer_name = row["Trainer"].split("(")[0].strip()
                except KeyError:
                    logging.warning("Failed to find trainer name for %s", url)

                place_str = str(row["Pla."]).strip()
                place = None
                if place_str.isnumeric():
                    place = int(place_str)

                jersey = str(row["Horse No."])

                handicap_weight = None
                try:
                    act_wt = str(row["Act. Wt."])
                    if act_wt != "---":
                        handicap_weight = float(act_wt) * 0.453592
                except KeyError:
                    logging.warning("Failed to find handicap weight for %s", url)

                horse_weight = None
                try:
                    horse_weight_str = row["Declar. Horse Wt."]
                    if horse_weight_str != "---":
                        horse_weight = float(horse_weight_str) * 0.453592
                except KeyError:
                    logging.warning("Failed to find horse weight for %s", url)

                starting_position = None
                starting_position_str = str(row["Dr."]).strip()
                if starting_position_str.strip() != "---":
                    starting_position = position_from_str(starting_position_str.strip())

                lbw: float | None = 0.0
                lbw_str = str(row["LBW"]).strip()
                lbw_str = lbw_str.split(" /")[0].strip()
                lbw_str = (
                    lbw_str.replace("/SHS", "")
                    .replace("5/H", "")
                    .replace("/HS", "")
                    .replace("/HD", "")
                    .replace("/NS", "")
                    .replace("/NK", "")
                    .replace("/H", "")
                    .replace("/SH", "")
                    .replace("/N", "")
                    .replace(" N", "")
                    .strip()
                )
                if lbw_str.endswith("/"):
                    lbw_str = lbw_str[:-1]
                if lbw_str in {"HD", "H", "HDS"}:
                    lbw = 2.4 * 0.2
                elif lbw_str in {"SH", "SHS", "SHSS", "HS", "S"}:
                    lbw = 2.4 * 0.2 * 0.5
                elif lbw_str in {"NOSE", "N", "NS"}:
                    lbw = 2.4 * 0.05
                elif lbw_str == "ML":
                    lbw = None
                elif lbw_str == "TO":
                    lbw = 100.0
                elif lbw_str in {"+NOSE", "+N"}:
                    lbw = 2.4 * 0.07
                elif lbw_str == "+SH":
                    lbw = 2.4 * 0.2 * 0.5
                elif lbw_str in {"HD+", "+HD"}:
                    lbw = 2.4 * 0.2 * 1.5
                elif lbw_str == "NK":
                    lbw = 2.4 * 0.4
                elif (
                    lbw_str
                    and lbw_str != "-"
                    and lbw_str != "---"
                    and lbw_str != "nan"
                    and lbw_str != "DNF"
                    and lbw_str != "M.L."
                    and lbw_str != "T.O."
                    and lbw_str != "U/R"
                    and lbw_str != "M. L."
                    and lbw_str != "TNP"
                    and lbw_str != "T.0."
                    and lbw_str != "D.N.F."
                    and lbw_str != "P.U."
                    and lbw_str != "U.R."
                    and lbw_str != "P U."
                    and lbw_str != "T 0."
                    and lbw_str != "3N"
                    and lbw_str != "4H"
                    and lbw_str != "SHDH"
                    and lbw_str != "NDH"
                ):
                    if lbw is not None:
                        if "-" in lbw_str:
                            lbw_horses, lbw_str = lbw_str.split("-")
                            lbw += 2.4 * int(lbw_horses)
                        if "/" in lbw_str:
                            lbw += 2.4 * (1.0 / int(lbw_str.split("/")[-1]))
                        else:
                            lbw += 2.4 * int(lbw_str.split()[0].strip())

                finish_time_str = row["Finish Time"].strip()
                end_dt = None
                if finish_time_str != "---":
                    end_t = datetime.datetime.strptime(finish_time_str, "%M:%S.%f")
                    end_dt = dt + datetime.timedelta(
                        minutes=end_t.minute,
                        seconds=end_t.second,
                        microseconds=end_t.microsecond,
                    )

                win_odds = str(row["Win Odds"]).strip()
                odds = None
                if win_odds != "---" and end_dt is not None:
                    odds = create_hkjc_hkjc_odds_model(float(win_odds), end_dt)

                horse_url = None
                jockey_url = None
                trainer_url = None
                for a in soup.find_all("a", href=True):
                    a_text = a.get_text().strip()
                    if a_text == horse_name:
                        a_url = urllib.parse.urljoin(url, a.get("href"))
                        o = urlparse(a_url)
                        if o.path.endswith("/Horse/Horse.aspx"):
                            horse_url = a_url
                    elif a_text == jockey_name:
                        a_url = urllib.parse.urljoin(url, a.get("href"))
                        o = urlparse(a_url)
                        if o.path.endswith("/Jockey/JockeyProfile.aspx"):
                            jockey_url = a_url
                    elif trainer_name is not None and a_text == trainer_name:
                        a_url = urllib.parse.urljoin(url, a.get("href"))
                        o = urlparse(a_url)
                        if o.path.endswith("/Trainers/TrainerProfile.aspx"):
                            trainer_url = a_url

                if place is None:
                    continue
                if horse_url is None:
                    logging.error(url)
                    logging.error(row)
                    raise ValueError("horse_url is null")

                team_model = create_hkjc_hkjc_team_model(
                    session=session,
                    horse_url=horse_url,
                    jockey_url=jockey_url,
                    trainer_url=trainer_url,
                    points=place,
                    jersey=jersey,
                    handicap_weight=handicap_weight,
                    horse_weight=horse_weight,
                    starting_position=starting_position,
                    lbw=lbw,
                    end_dt=end_dt,
                    odds=[] if odds is None else [odds],
                )
                teams.append(team_model)
        elif count == 3:
            for _, row in df.iterrows():
                row_df = row.to_frame()
                pool = str(row_df.iat[0, 0]).strip()

                try:
                    dividend_str = str(row_df.iat[2, 0]).strip().lower()
                    if dividend_str == "details":
                        continue
                    if dividend_str == "refund":
                        continue
                    if dividend_str == "not win":
                        continue
                    if dividend_str == "detail":
                        continue
                    dividend_str = (
                        dividend_str.split("/", maxsplit=1)[0]
                        .replace(",", "")
                        .replace("$", "")
                        .replace("(", "")
                        .replace(")", "")
                    )
                    if dividend_str == "win":
                        continue
                except IndexError:
                    continue
                dividend = float(dividend_str)

                combination = None
                next_td = False
                for td in soup.find_all("td"):
                    td_text = td.get_text().strip()
                    if td_text == pool:
                        next_td = True
                    else:
                        if next_td:
                            combination = []
                            for combo in td_text.split(","):
                                for sub_combo in combo.split("/"):
                                    combination.append(sub_combo)
                            break
                        next_td = False

                if combination is None:
                    continue

                dividend_model = create_hkjc_hkjc_dividend_model(
                    pool=pool,
                    combination=combination,
                    team_models=teams,
                    dividend=float(dividend),
                )
                dividends.append(dividend_model)

    if race_track is None:
        logging.error(url)
        raise ValueError("race_track is null")

    end_dts = [x.end_dt for x in teams if x.end_dt is not None]
    end_dt = None if not end_dts else min(end_dts)

    return GameModel(
        dt=dt,
        week=None,
        game_number=game_number,
        venue=create_hkjc_hkjc_venue_model(
            session=session,
            dt=dt,
            venue_code=venue_code,
            race_track=race_track,
            version=VENUE_VERSION,
        ),
        teams=teams,
        end_dt=end_dt,
        attendance=None,
        league=str(League.HKJC),
        year=dt.year,
        season_type=None,
        postponed=None,
        play_off=None,
        distance=distance,
        dividends=dividends,
        pot=pot,
        version=version,
        umpires=[],
        best_of=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_hkjc_hkjc_game_model(
    session: ScrapeSession,
    html: str,
    url: str,
    version: str,
) -> GameModel | None:
    """Create a game model from NBA API."""
    return _create_hkjc_hkjc_game_model(
        session=session, html=html, url=url, version=version
    )


def create_hkjc_hkjc_game_model(
    session: ScrapeSession,
    html: str,
    url: str,
) -> GameModel | None:
    """Create a game model from NBA API."""
    if not pytest_is_running.is_running():
        return _cached_create_hkjc_hkjc_game_model(
            session=session, html=html, url=url, version=VERSION
        )
    with session.cache_disabled():
        return _cached_create_hkjc_hkjc_game_model(
            session=session, html=html, url=url, version=VERSION
        )



"""AFL AFL bookie model."""

from ...bookie_model import BookieModel


def create_afl_afl_bookie_model() -> BookieModel:
    """Create bookie model from AFL.com."""
    return BookieModel(identifier="sportsbet", name="sportsbet")



"""Tests for the NCAAB sportsdb league model class."""
import unittest

import requests_cache
from sportsball.data.ncaab.sportsreference.ncaab_sportsreference_league_model import NCAABSportsReferenceLeagueModel
from sportsball.data.league import League


class TestNCAABSportsReferenceLeagueModel(unittest.TestCase):

    def setUp(self):
        self.session = requests_cache.CachedSession(backend="memory")
        self.model = NCAABSportsReferenceLeagueModel(self.session)

    def test_league(self):
        self.assertEqual(self.model.league, League.NCAAB)



"""WNBA combined league model."""

# pylint: disable=line-too-long
from scrapesession.scrapesession import ScrapeSession  # type: ignore

from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.wnba_espn_league_model import WNBAESPNLeagueModel
from ..oddsportal.wnba_oddsportal_league_model import WNBAOddsPortalLeagueModel
from ..wnbacom.wnba_wnbacom_league_model import WNBAWNBAComLeagueModel

MERCURY = "11"
LIBERTY = "9"
MYSTICS = "16"
ROCKERS = "2"
MONARCHS = "13"
SPARKS = "6"
COMETS = "4"
STARZZ = "15"
LYNX = "8"
STING = "1"
MIRACLE = "10"
SHOCK = "3"
SOL = "7"
STORM = "14"
FIRE = "12"
FEVER = "5"
WEST = "99"
EAST = "98"
SUN = "18"
SILVER_STARS = "17"
TEAM_USA = "97"
TEAM_WNBA = "96"
SKY = "19"
DREAM = "20"
TEAM_CHINA = "21"
TEAM_POLAND = "22"
TEAM_GREAT_BRITAIN = "24"
JX_SUNFLOWERS = "23"
TEAM_JAPAN = "25"
TEAM_BRAZIL = "17473"
TEAM_AUSTRALIA = "17472"
TEAM_PARKER = "111720"
TEAM_DELLE_DONNE = "111719"
TEAM_WILSON = "112530"
VALKYRIES = "129689"
TEAM_COLLIER = "131247"
TEAM_CLARK = "131246"
PUERTO_RICO = "104959"
TEAM_STEWART = "126287"
ANTELOPES = "130927"
WNBA_TEAM_IDENTITY_MAP: dict[str, str] = {
    # ESPN
    "11": MERCURY,
    "9": LIBERTY,
    "16": MYSTICS,
    "2": ROCKERS,
    "13": MONARCHS,
    "6": SPARKS,
    "4": COMETS,
    "15": STARZZ,
    "8": LYNX,
    "1": STING,
    "10": MIRACLE,
    "3": SHOCK,
    "7": SOL,
    "14": STORM,
    "12": FIRE,
    "5": FEVER,
    "99": WEST,
    "98": EAST,
    "18": SUN,
    "17": SILVER_STARS,
    "97": TEAM_USA,
    "96": TEAM_WNBA,
    "19": SKY,
    "20": DREAM,
    "21": TEAM_CHINA,
    "22": TEAM_POLAND,
    "24": TEAM_GREAT_BRITAIN,
    "23": JX_SUNFLOWERS,
    "25": TEAM_JAPAN,
    # OddsPortal
    "17473": TEAM_BRAZIL,
    "17475": TEAM_JAPAN,
    "17472": TEAM_AUSTRALIA,
    "111720": TEAM_PARKER,
    "111719": TEAM_DELLE_DONNE,
    "112530": TEAM_WILSON,
    "129689": VALKYRIES,
    "131247": TEAM_COLLIER,
    "131246": TEAM_CLARK,
    "104959": PUERTO_RICO,
    "126287": TEAM_STEWART,
    "130927": ANTELOPES,
}
MADISON_SQUARE_GARDEN = "1830"
CAPITAL_ONE_ARENA = "1823"
QUICKEN_LOANS_ARENA = "3417"
CRYPTO_COM_ARENA = "1841"
TARGET_CENTER = "2029"
THE_PALACE_OF_AUBURN_HILLS = "3419"
KEY_ARENA = "2184"
AMERICAN_AIRLINES_ARENA = "3438"
PHX_ARENA = "1949"
ROSE_GARDEN = "3423"
MOHEGAN_SUN_ARENA = "2160"
FROST_BANK_CENTER = "780"
PRUDENTIAL_CENTER = "1826"
BOK_CENTER = "1911"
GAINBRIDGE_FIELDHOUSE = "2183"
STATE_FARM_ARENA = "1827"
ALLSTATE_ARENA = "977"
THE_ARENA_AT_GWINNETT_CENTER = "1245"
JACOBY_DICKENS_CENTER = "4332"
WALTER_PYRAMID = "539"
SRC_ARENA = "2082"
BISMARCK_EVENT_CENTER = "3446"
MCCAMISH_PAVILION = "2206"
HP_FIELD_HOUSE = "2222"
LAKIN_CENTER = "3500"
BOB_CARPENTER_CENTER = "2189"
FREEMAN_COLISEUM = "2935"
CORTEVA_COLISEUM = "4947"
KFC_YUM_CENTER = "2030"
LEVIEN_GYMNASIUM = "2002"
MAYO_CIVIC_CENTER = "5240"
HUTTO_PATTERSON_GYMNASIUM = "3498"
WILLIAMS_ARENA = "2094"
GRAND_CASINO_ARENA = "1835"
COLLEGE_PARK_CENTER = "3416"
DESERT_FINANCIAL_ARENA = "833"
CHARLES_E_SMITH_CENTER = "2125"
EAGLEBANK_ARENA = "2167"
MICHELOB_ULTRA_ARENA = "4005"
WESTCHESTER_COUNTY_CENTER = "5584"
WINTRUST_ARENA = "5427"
WELLS_FARGO_ARENA = "1475"
TOTAL_MORTGAGE_ARENA = "1259"
ANGEL_OF_THE_WINDS_ARENA = "5242"
CAREFIRST_ARENA = "5946"
THOMAS_AND_MACK_CENTER = "2083"
ALASKA_AIRLINES_ARENA = "782"
BARCLAYS_CENTER = "3559"
T_MOBILE_ARENA = "5060"
ALBANY_CIVIC_CENTER = "6178"
COX_PAVILION = "1051"
MVP_ARENA = "20"
IMG_ACADEMY = "7097"
GATEWAY_CENTER = "6282"
SAP_CENTER_AT_SAN_JOSE = "1653"
CLIMATE_PLEDGE_ARENA = "7224"
CFG_BANK_ARENA = "4938"
CHASE_CENTER = "6270"
ROGERS_ARENA = "1833"
AMERICAN_AIRLINES_CENTER = "516"
UNITED_CENTER = "1847"
TD_GARDEN = "1824"
UBS_ARENA = "7226"
COLONIAL_LIFE_ARENA = "1962"
ROGERS_PLACE = "5282"
SCOTIABANK_ARENA = "1838"
HINKLE_FIELDHOUSE = "2032"
GLOBAL_CREDIT_UNION_ARENA = "4534"
LOS_ANGELES_CONVENTION_CENTER = "2512"
MATTHEW_KNIGHT_ARENA = "1977"
CARVER_HAWKEYE_ARENA = "363"
PETE_MAROVICH_ASSEMBLY_CENTER = "557"
PURCELL_PAVILION = "282"
WNBA_VENUE_IDENTITY_MAP: dict[str, str] = {
    # ESPN
    "1830": MADISON_SQUARE_GARDEN,
    "1823": CAPITAL_ONE_ARENA,
    "3417": QUICKEN_LOANS_ARENA,
    "1841": CRYPTO_COM_ARENA,
    "2029": TARGET_CENTER,
    "3419": THE_PALACE_OF_AUBURN_HILLS,
    "2184": KEY_ARENA,
    "3438": AMERICAN_AIRLINES_ARENA,
    "1949": PHX_ARENA,
    "3423": ROSE_GARDEN,
    "2160": MOHEGAN_SUN_ARENA,
    "780": FROST_BANK_CENTER,
    "1826": PRUDENTIAL_CENTER,
    "1911": BOK_CENTER,
    "2183": GAINBRIDGE_FIELDHOUSE,
    "1827": STATE_FARM_ARENA,
    "977": ALLSTATE_ARENA,
    "1245": THE_ARENA_AT_GWINNETT_CENTER,
    "4332": JACOBY_DICKENS_CENTER,
    "539": WALTER_PYRAMID,
    "2082": SRC_ARENA,
    "3446": BISMARCK_EVENT_CENTER,
    "2206": MCCAMISH_PAVILION,
    "2222": HP_FIELD_HOUSE,
    "3500": LAKIN_CENTER,
    "2189": BOB_CARPENTER_CENTER,
    "2935": FREEMAN_COLISEUM,
    "4947": CORTEVA_COLISEUM,
    "2030": KFC_YUM_CENTER,
    "2002": LEVIEN_GYMNASIUM,
    "5240": MAYO_CIVIC_CENTER,
    "3498": HUTTO_PATTERSON_GYMNASIUM,
    "2094": WILLIAMS_ARENA,
    "1835": GRAND_CASINO_ARENA,
    "3416": COLLEGE_PARK_CENTER,
    "833": DESERT_FINANCIAL_ARENA,
    "2125": CHARLES_E_SMITH_CENTER,
    "2167": EAGLEBANK_ARENA,
    "4005": MICHELOB_ULTRA_ARENA,
    "5584": WESTCHESTER_COUNTY_CENTER,
    "5427": WINTRUST_ARENA,
    "1475": WELLS_FARGO_ARENA,
    "1259": TOTAL_MORTGAGE_ARENA,
    "5242": ANGEL_OF_THE_WINDS_ARENA,
    "5946": CAREFIRST_ARENA,
    "2083": THOMAS_AND_MACK_CENTER,
    "782": ALASKA_AIRLINES_ARENA,
    "3559": BARCLAYS_CENTER,
    "5060": T_MOBILE_ARENA,
    "6178": ALBANY_CIVIC_CENTER,
    "1051": COX_PAVILION,
    "20": MVP_ARENA,
    "7097": IMG_ACADEMY,
    "6282": GATEWAY_CENTER,
    "1653": SAP_CENTER_AT_SAN_JOSE,
    "7224": CLIMATE_PLEDGE_ARENA,
    "4938": CFG_BANK_ARENA,
    "6270": CHASE_CENTER,
    "1833": ROGERS_ARENA,
    "516": AMERICAN_AIRLINES_CENTER,
    "1847": UNITED_CENTER,
    "1824": TD_GARDEN,
    "7226": UBS_ARENA,
    "1962": COLONIAL_LIFE_ARENA,
    "5282": ROGERS_PLACE,
    "1838": SCOTIABANK_ARENA,
    "2032": HINKLE_FIELDHOUSE,
    "4534": GLOBAL_CREDIT_UNION_ARENA,
    "2512": LOS_ANGELES_CONVENTION_CENTER,
    "1977": MATTHEW_KNIGHT_ARENA,
    "363": CARVER_HAWKEYE_ARENA,
    "557": PETE_MAROVICH_ASSEMBLY_CENTER,
    "282": PURCELL_PAVILION,
}
WNBA_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class WNBACombinedLeagueModel(CombinedLeagueModel):
    """WNBA combined implementation of the league model."""

    def __init__(self, session: ScrapeSession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.WNBA,
            [
                WNBAESPNLeagueModel(session, position=0),
                WNBAOddsPortalLeagueModel(session, position=1),
                WNBAWNBAComLeagueModel(session, position=2),
            ],
            league_filter,
        )

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return WNBA_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return WNBA_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return WNBA_PLAYER_IDENTITY_MAP

    @classmethod
    def name(cls) -> str:
        return "wnba-combined-league-model"
